#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <stdio.h>

// TILE_SIZE defines the block dimensions for matrix multiplication tiling
constexpr int TILE_SIZE = 16;

// Linear Forward Kernel with shared memory bias caching
// Computes output = x * weight^T + bias
// x: [batch_size, in_features]
// weight: [out_features, in_features]
// bias: [out_features]
// output: [batch_size, out_features]

template <typename scalar_t, int TILE_SIZE>
__global__ void linear_forward_kernel_optimized(
    const scalar_t* __restrict__ x,
    const scalar_t* __restrict__ weight,
    const scalar_t* __restrict__ bias,
    scalar_t* __restrict__ output,
    int batch_size,
    int in_features,
    int out_features) {

  int row = blockIdx.y * TILE_SIZE + threadIdx.y; // batch index
  int col = blockIdx.x * TILE_SIZE + threadIdx.x; // output feature index

  __shared__ scalar_t A_tile[TILE_SIZE][TILE_SIZE];
  __shared__ scalar_t B_tile[TILE_SIZE][TILE_SIZE];
  __shared__ scalar_t bias_tile[TILE_SIZE];  // Shared bias for this block's column tile

  // Load bias values for the columns in this block once
  if (threadIdx.y == 0) {
    int bias_idx = blockIdx.x * TILE_SIZE + threadIdx.x;
    bias_tile[threadIdx.x] = (bias_idx < out_features) ? bias[bias_idx] : static_cast<scalar_t>(0);
  }
  __syncthreads();

  scalar_t sum = 0;
  int numTiles = (in_features + TILE_SIZE - 1) / TILE_SIZE;

  for (int t = 0; t < numTiles; t++) {
    int a_col = t * TILE_SIZE + threadIdx.x;
    A_tile[threadIdx.y][threadIdx.x] = (row < batch_size && a_col < in_features) 
                                         ? x[row * in_features + a_col] 
                                         : static_cast<scalar_t>(0);

    int k = t * TILE_SIZE + threadIdx.y;
    B_tile[threadIdx.y][threadIdx.x] = (col < out_features && k < in_features) 
                                         ? weight[col * in_features + k] 
                                         : static_cast<scalar_t>(0);
    __syncthreads();

    #pragma unroll
    for (int i = 0; i < TILE_SIZE; i++) {
      if (t * TILE_SIZE + i < in_features) {
        sum += A_tile[threadIdx.y][i] * B_tile[i][threadIdx.x];
      }
    }
    __syncthreads();
  }

  if (row < batch_size && col < out_features) {
    output[row * out_features + col] = sum + bias_tile[threadIdx.x];
  }
}

// Group Normalization Kernel with shared memory caching
// Each block processes one (batch, group) pair.
// x: [batch_size, num_channels]
// gamma, beta: [num_channels] used for scaling and shifting
// num_groups divides num_channels evenly.

template <typename scalar_t>
__global__ void group_norm_forward_kernel_shared(
    const scalar_t* __restrict__ x,
    const scalar_t* __restrict__ gamma,
    const scalar_t* __restrict__ beta,
    scalar_t* __restrict__ output,
    int batch_size,
    int num_channels,
    int num_groups) {

  int channels_per_group = num_channels / num_groups;
  int idx = blockIdx.x; // Each block corresponds to one (batch, group) pair
  int batch = idx / num_groups;
  int group = idx % num_groups;

  // Allocate shared memory for group data and a reduction buffer
  // The first part holds the group's data; the second part is used for reduction.
  extern __shared__ char shared_mem[];
  scalar_t* group_data = reinterpret_cast<scalar_t*>(shared_mem);
  scalar_t* reduction = reinterpret_cast<scalar_t*>(shared_mem + channels_per_group * sizeof(scalar_t));

  // Load the entire group's data from global memory into shared memory
  for (int i = threadIdx.x; i < channels_per_group; i += blockDim.x) {
    int channel = group * channels_per_group + i;
    group_data[i] = x[batch * num_channels + channel];
  }
  __syncthreads();

  // Compute the mean using a reduction in shared memory
  scalar_t local_sum = 0;
  for (int i = threadIdx.x; i < channels_per_group; i += blockDim.x) {
    local_sum += group_data[i];
  }
  reduction[threadIdx.x] = local_sum;
  __syncthreads();
  for (int s = blockDim.x >> 1; s > 0; s >>= 1) {
    if (threadIdx.x < s) {
      reduction[threadIdx.x] += reduction[threadIdx.x + s];
    }
    __syncthreads();
  }
  scalar_t mean = reduction[0] / channels_per_group;
  __syncthreads();

  // Compute the variance using the shared group data
  scalar_t local_var = 0;
  for (int i = threadIdx.x; i < channels_per_group; i += blockDim.x) {
    scalar_t diff = group_data[i] - mean;
    local_var += diff * diff;
  }
  reduction[threadIdx.x] = local_var;
  __syncthreads();
  for (int s = blockDim.x >> 1; s > 0; s >>= 1) {
    if (threadIdx.x < s) {
      reduction[threadIdx.x] += reduction[threadIdx.x + s];
    }
    __syncthreads();
  }
  scalar_t var = reduction[0] / channels_per_group;
  scalar_t inv_std = rsqrtf(var + 1e-5f);
  __syncthreads();

  // Normalize the group data and apply scale (gamma) and shift (beta)
  for (int i = threadIdx.x; i < channels_per_group; i += blockDim.x) {
    int channel = group * channels_per_group + i;
    scalar_t norm = (group_data[i] - mean) * inv_std;
    output[batch * num_channels + channel] = norm * gamma[channel] + beta[channel];
  }
}

// Hardtanh Activation Kernel with grid-stride loop
// Applies activation: output = min(max(x, min_val), max_val)

template <typename scalar_t>
__global__ void hardtanh_forward_kernel(
    const scalar_t* __restrict__ x,
    scalar_t min_val,
    scalar_t max_val,
    scalar_t* __restrict__ output,
    size_t total_elements) {
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;
  for (; idx < total_elements; idx += stride) {
    scalar_t val = x[idx];
    if (val < min_val) val = min_val;
    if (val > max_val) val = max_val;
    output[idx] = val;
  }
}

// Host function to launch the linear kernel
void linear_forward_cuda_optimized(
    at::Tensor x, 
    at::Tensor weight, 
    at::Tensor bias, 
    at::Tensor output) {
  const int batch_size = x.size(0);
  const int in_features = x.size(1);
  const int out_features = weight.size(0);

  dim3 block(TILE_SIZE, TILE_SIZE);
  dim3 grid((out_features + TILE_SIZE - 1) / TILE_SIZE,
            (batch_size + TILE_SIZE - 1) / TILE_SIZE);

  AT_DISPATCH_FLOATING_TYPES(x.scalar_type(), "linear_forward_cuda_optimized", ([&] {
    linear_forward_kernel_optimized<scalar_t, TILE_SIZE><<<grid, block>>>(
        x.data_ptr<scalar_t>(),
        weight.data_ptr<scalar_t>(),
        bias.data_ptr<scalar_t>(),
        output.data_ptr<scalar_t>(),
        batch_size,
        in_features,
        out_features);
  }));

  cudaError_t err = cudaGetLastError();
  if (err != cudaSuccess) {
    printf("Error in linear_forward_cuda_optimized: %s\n", cudaGetErrorString(err));
  }
}

// Host function to launch the group normalization kernel using shared memory
void group_norm_forward_cuda_shared(
    at::Tensor x, 
    at::Tensor gamma, 
    at::Tensor beta, 
    int64_t num_groups, 
    at::Tensor output) {
  const int batch_size = x.size(0);
  const int num_channels = x.size(1);
  int channels_per_group = num_channels / num_groups;

  int total_blocks = batch_size * num_groups;
  int threads = (channels_per_group < 256) ? channels_per_group : 256;
  
  // Shared memory: group_data (channels_per_group elements) + reduction (threads elements)
  size_t shared_mem = channels_per_group * sizeof(float) + threads * sizeof(float);

  AT_DISPATCH_FLOATING_TYPES(x.scalar_type(), "group_norm_forward_cuda_shared", ([&] {
    group_norm_forward_kernel_shared<scalar_t><<<total_blocks, threads, shared_mem>>>(
        x.data_ptr<scalar_t>(),
        gamma.data_ptr<scalar_t>(),
        beta.data_ptr<scalar_t>(),
        output.data_ptr<scalar_t>(),
        batch_size,
        num_channels,
        num_groups);
  }));

  cudaError_t err = cudaGetLastError();
  if (err != cudaSuccess) {
    printf("Error in group_norm_forward_cuda_shared: %s\n", cudaGetErrorString(err));
  }
}

// Host function to launch the hardtanh kernel
void hardtanh_forward_cuda(
    at::Tensor x, 
    float min_val, 
    float max_val,
    at::Tensor output) {
  size_t total_elements = x.numel();
  int threads = 256;
  int blocks = (total_elements + threads - 1) / threads;

  AT_DISPATCH_FLOATING_TYPES(x.scalar_type(), "hardtanh_forward_cuda", ([&] {
    hardtanh_forward_kernel<scalar_t><<<blocks, threads>>>(
        x.data_ptr<scalar_t>(),
        static_cast<scalar_t>(min_val),
        static_cast<scalar_t>(max_val),
        output.data_ptr<scalar_t>(),
        total_elements);
  }));

  cudaError_t err = cudaGetLastError();
  if (err != cudaSuccess) {
    printf("Error in hardtanh_forward_cuda: %s\n", cudaGetErrorString(err));
  }
}

// Combined host function that runs linear, group norm, and hardtanh sequentially
at::Tensor module_fn_cuda_forward(
    at::Tensor x,
    at::Tensor weight,
    at::Tensor bias,
    at::Tensor group_norm_weight,
    at::Tensor group_norm_bias,
    int64_t num_groups,
    float hardtanh_min,
    float hardtanh_max) {
  
  // Ensure inputs are contiguous
  x = x.contiguous();
  weight = weight.contiguous();
  bias = bias.contiguous();
  group_norm_weight = group_norm_weight.contiguous();
  group_norm_bias = group_norm_bias.contiguous();

  int64_t batch_size = x.size(0);
  int64_t in_features = x.size(1);
  int64_t out_features = weight.size(0);

  auto options = x.options();
  at::Tensor linear_output = at::empty({batch_size, out_features}, options);
  at::Tensor group_norm_output = at::empty({batch_size, out_features}, options);
  at::Tensor output = at::empty({batch_size, out_features}, options);

  // Run the linear layer with shared memory bias caching
  linear_forward_cuda_optimized(x, weight, bias, linear_output);

  // Run group normalization with shared memory caching of group data
  group_norm_forward_cuda_shared(linear_output, group_norm_weight, group_norm_bias, num_groups, group_norm_output);

  // Apply hardtanh activation
  hardtanh_forward_cuda(group_norm_output, hardtanh_min, hardtanh_max, output);

  return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
  m.def("forward", &module_fn_cuda_forward, "Forward pass (CUDA shared memory optimized)");
}
